(***********************************************************************)
(*                                                                     *)
(*                              OCamlGen                               *)
(*                                                                     *)
(*                       OCaml module Generator                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 1997-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: og_lexer.mll,v 1.8 2014/01/31 09:51:52 fclement Exp $ *)

(* The lexer for data type definitions *)

{
open Og_types;;

type error =
  | Illegal_character
  | Unterminated_comment
  | Unterminated_string
;;

exception Error of error * int * int;;

(* For nested comments *)

let comment_depth = ref 0;;

(* The table of keywords *)

let keyword_table =
  Og_misc.hash_table_make 13 [

    "of", OF;

    "exception", EXCEPTION;
    "external", EXTERNAL;

    "false", FALSE;

    "mod", MOD;
    "mutable", MUTABLE;

    "private", PRIVATE;

    "true", TRUE;

    "type", TYPE;
    "and", AND;

    "open", OPEN;
    "include", INCLUDE;

    "module", MODULE;
    "sig", SIG;
    "end", END;

    "val", VAL;
]
;;

(* To buffer string literals *)

let initial_string_buffer = String.create 256;;
let string_buff = ref initial_string_buffer;;
let string_index = ref 0;;

let reset_string_buffer () =
  string_buff := initial_string_buffer;
  string_index := 0;
;;

let store_string_char c =
  if !string_index >= String.length (!string_buff) then begin
    let new_buff = String.create (String.length (!string_buff) * 2) in
    String.blit (!string_buff) 0 new_buff 0 (String.length (!string_buff));
    string_buff := new_buff;
  end;
  String.unsafe_set (!string_buff) (!string_index) c;
  incr string_index
;;

let get_stored_string () =
  let s = String.sub (!string_buff) 0 (!string_index) in
  string_buff := initial_string_buffer;
  s
;;

let comment_buffer = Buffer.create 13;;
let reset_comment_buffer () = Buffer.clear comment_buffer;;

let store_comment_char c = Buffer.add_char comment_buffer c;;
let store_comment_string s = Buffer.add_string comment_buffer s;;

let get_stored_comment () = Buffer.contents comment_buffer;;

(* To translate escape sequences *)

let char_for_backslash = function
  | 'n' -> '\010'
  | 'r' -> '\013'
  | 'b' -> '\008'
  | 't' -> '\009'
  | c   -> c
;;

let char_for_decimal_code lexbuf i =
  let c =
    100 * (Char.code (Lexing.lexeme_char lexbuf i) - 48) +
     10 * (Char.code (Lexing.lexeme_char lexbuf (i + 1)) - 48) +
          (Char.code (Lexing.lexeme_char lexbuf (i + 2)) - 48) in
  Char.chr (c land 0xFF)
;;

(* To store the position of the beginning of a string or comment *)

let start_pos = ref 0;;

(* Error report *)

let report_error = function
    Illegal_character ->
      Format.print_string "Illegal character"
  | Unterminated_comment ->
      Format.print_string "Comment not terminated"
  | Unterminated_string ->
      Format.print_string "String literal not terminated"
;;

let token_of_symbolic s =
  assert (String.length s <> 0);
  match s with
  | "*"  -> STAR
  | "?"  -> QUESTION
  | "<"  -> LESS
  | "="  -> EQUAL
  | "|"  -> BAR
  | ">"  -> GREATER
  | "-"  -> MINUS
  | "->" -> MINUSGREATER
  | "<>" -> LESSGREATER
  | "<=" -> LESSEQUAL
  | "&&" -> AMPERAMPER
  | ">=" -> GREATEREQUAL
  | "!=" -> BANGEQUAL
  | "==" -> EQUALEQUAL
  | "||" -> BARBAR
  | _ ->
    match s.[0] with
    | '~' -> TILDA_OP s
    | '!' -> BANG_OP s
    | '@' -> AT_OP s
    | '%' -> PERCENT_OP s
    | '^' -> HAT_OP s
    | '&' -> AMPER_OP s
    | '*' -> STAR_OP s
    | '-' -> MINUS_OP s
    | '=' -> EQUAL_OP s
    | '+' -> PLUS_OP s
    | '<' -> LESS_OP s
    | '>' -> GREATER_OP s
    | '/' -> SLASH_OP s
    | '?' -> QUESTION_OP s
    | _ -> assert false
;;

}

let space = [ ' ' '\010' '\013' '\009' '\012' ]

let symbolic = [ '~' '!' '@' '%' '^' '&' '*' '-' '=' '+' '|' '<' '>' '/' '?' ]

let suite_symbolic = symbolic | '_'

let symbol = symbolic suite_symbolic*

rule token = parse
  | space +
      { token lexbuf }
  | [ '_' 'a'-'z' '\223'-'\246' '\248'-'\255' ]
    ([ 'A'-'Z' 'a'-'z' '_' '\192'-'\214' '\216'-'\246' '\248'-'\255'
      '\'' '0'-'9' ]) *
      { let s = Lexing.lexeme lexbuf in
          try
            Hashtbl.find keyword_table s
          with Not_found ->
            LIDENT s }
  | [ 'A'-'Z' '\192'-'\214' '\216'-'\222' ]
    ([ 'A'-'Z' 'a'-'z' '_' '\192'-'\214' '\216'-'\246' '\248'-'\255'
      '\'' '0'-'9' ]) *
      { UIDENT (Lexing.lexeme lexbuf) }       (* No capitalized keywords *)
  | ['0'-'9']+
    | '0' ['x' 'X'] ['0'-'9' 'A'-'F' 'a'-'f']+
    | '0' ['o' 'O'] ['0'-'7']+
    | '0' ['b' 'B'] ['0'-'1']+
      { INT (int_of_string(Lexing.lexeme lexbuf)) }
  | ['0'-'9']+ ('.' ['0'-'9']*)? (['e' 'E'] ['+' '-']? ['0'-'9']+)?
      { FLOAT (Lexing.lexeme lexbuf) }
  | "\""
      { reset_string_buffer ();
        let string_start = Lexing.lexeme_start lexbuf in
        start_pos := string_start;
        string lexbuf;
        lexbuf.Lexing.lex_start_pos <-
          string_start - lexbuf.Lexing.lex_abs_pos;
        STRING (get_stored_string ()) }
  | "'" [^ '\\' '\''] "'"
      { CHAR (Lexing.lexeme_char lexbuf 1) }
  | "'" '\\' ['\\' '\'' 'n' 't' 'b' 'r'] "'"
      { CHAR (char_for_backslash (Lexing.lexeme_char lexbuf 2)) }
  | "'" '\\' ['0'-'9'] ['0'-'9'] ['0'-'9'] "'"
      { CHAR (char_for_decimal_code lexbuf 2) }
  | "\'" { QUOTE }
  | "(*"
      { comment_depth := 1;
        reset_comment_buffer ();
        store_comment_string (Lexing.lexeme lexbuf);
        comment lexbuf;
        COMMENT (get_stored_comment ()) }
  | "#" [' ' '\t']* ['0'-'9']+ [' ' '\t']* "\"" [^ '\n' '\r'] *
    ('\n' | '\r' | "\r\n")
      (* # linenum "filename" flags \n *)
      { token lexbuf }
  | "("  { LPAREN }
  | ")"  { RPAREN }
  | "["  { LBRACKET }
  | "]"  { RBRACKET }
  | "{"  { LBRACE }
  | "}"  { RBRACE }
  | "[|" { LBRACKETBAR }
  | "|]" { BARRBRACKET }
  | "."  { DOT }
  | ".." { DOTDOT }
  | ","  { COMMA }
  | ":"  { COLON }
  | ";"  { SEMI }
  | ";;" { SEMISEMI }
  | symbol { token_of_symbolic (Lexing.lexeme lexbuf) }
  | eof { EOF }
  | _
      { raise (Error (Illegal_character,
                      Lexing.lexeme_start lexbuf, Lexing.lexeme_end lexbuf)) }

and comment = parse
  | "(*"
      { store_comment_string (Lexing.lexeme lexbuf);
        comment_depth := succ !comment_depth;
        comment lexbuf }
  | "*)" space*
      { comment_depth := pred !comment_depth;
        store_comment_string (Lexing.lexeme lexbuf);
        if !comment_depth > 0 then comment lexbuf }
  | "\""
      { store_comment_char '"';
        reset_string_buffer ();
        start_pos := Lexing.lexeme_start lexbuf;
        string lexbuf;
        store_comment_string (get_stored_string ());
        store_comment_char '"';
        string_buff := initial_string_buffer;
        comment lexbuf }
  | "''"
      { store_comment_string (Lexing.lexeme lexbuf); comment lexbuf }
  | "'" [^ '\\' '\''] "'"
      { store_comment_string (Lexing.lexeme lexbuf); comment lexbuf }
  | "'\\" ['\\' '\'' 'n' 't' 'b' 'r'] "'"
      { store_comment_string (Lexing.lexeme lexbuf); comment lexbuf }
  | "'\\" ['0'-'9'] ['0'-'9'] ['0'-'9'] "'"
      { store_comment_string (Lexing.lexeme lexbuf); comment lexbuf }
  | eof
      { raise (Error (Unterminated_comment, !start_pos, !start_pos+2)) }
  | _
      { store_comment_char (Lexing.lexeme_char lexbuf 0);
        comment lexbuf }

and string = parse
  | '"'
      { () }
  | '\\' ("\010" | "\013" | "\013\010") [' ' '\009'] *
      { string lexbuf }
  | '\\' ['\\' '"' 'n' 't' 'b' 'r']
      { store_string_char(char_for_backslash (Lexing.lexeme_char lexbuf 1));
        string lexbuf }
  | '\\' ['0'-'9'] ['0'-'9'] ['0'-'9']
      { store_string_char(char_for_decimal_code lexbuf 1);
         string lexbuf }
  | eof
      { raise (Error (Unterminated_string, !start_pos, !start_pos+1)) }
  | _
      { store_string_char (Lexing.lexeme_char lexbuf 0);
        string lexbuf }
